-- RBT-TEST - ДПРД для столбцов и строк
-- Проверка возможности ограничения доступа к столбцам и строкам
-- Права доступа для столбцов SELECT, INSERT, UPDATE, REFERENCES
--
\i 'support/init.sql'

------------------------------------------------------------
-- I Проверка возможности ограничения доступа к колонке
------------------------------------------------------------
------------------------------------------------------------
-- 1. Подготовка
------------------------------------------------------------
-- Подключаемся суперпользователем
\c - postgres
-- Создадим таблицу для проверки. Колонки for_all_key и for_all_value смогут читать все, колонку for_user_u сможет читать только пользователь u_0_00.
CREATE TABLE test1 (for_all_key integer PRIMARY KEY, for_all_value text, for_user_u text);
-- Добавим данные в таблицу test1
INSERT INTO test1 VALUES (1,'its for all', 'its only for user u_0_00'), (2,'its for all', 'its only for user u_0_00'), (3,'its for all', 'its only for user u_0_00');
-- Выдадим привилегию чтения на колонку for_user_u пользователю u_0_00.
GRANT SELECT (for_user_u) ON TABLE test1 TO u_0_00;
GRANT ALL PRIVILEGES (for_all_key) ON TABLE test1 TO public;
GRANT ALL PRIVILEGES (for_all_value) ON TABLE test1 TO public;
------------------------------------------------------------
-- 2. Проверка ограничения доступа на чтение
------------------------------------------------------------
-- Подключаемся пользователем u_0_00
\c - u_0_00
-- Пользователь u_0_00 может получить все данные из таблицы test1
SELECT * FROM test1;
-- Подключаемся пользователем u1_0_00
\c - u1_0_00
-- Пользователь u1_0_00 может получить доступ на чтение только к полям for_all_key и for_all_value
SELECT for_all_key FROM test1;
SELECT for_all_value FROM test1;
SELECT for_all_key, for_all_value FROM test1;
SELECT for_all_key, for_all_value, for_user_u FROM test1;
SELECT * FROM test1;
------------------------------------------------------------
-- 3. Проверка ограничения создания
------------------------------------------------------------
-- Подключаемся пользователем u_0_00
\c - u_0_00
-- Пользователю u_0_00 не разрешено добавлять данные в колонку for_user_u
INSERT INTO test1 (for_all_key, for_user_u) VALUES (4, 'test_value');
-- Тогда как добавлять данные в колонки for_all_key, for_all_value разрешено
INSERT INTO test1 (for_all_key, for_all_value) VALUES (4, 'test_value');
------------------------------------------------------------
-- 4. Проверка ограничения модификации
------------------------------------------------------------
-- Пользователю u_0_00 не разрешено изменять данные в колонке for_user_u (модификация)
UPDATE test1 SET for_user_u = 'its for all + test_value' WHERE for_all_key = 1;
-- Тогда как модификация колонки for_all_value разрешена
UPDATE test1 SET for_all_value = 'its for all + test_value' WHERE for_all_key = 1;
------------------------------------------------------------
-- 5. Проверка ограничения удаления
------------------------------------------------------------
-- Пользователю не разрешено удалять данные из колонки for_user_u
DELETE FROM test1 WHERE for_all_key = 1;
------------------------------------------------------------
-- II Проверка возможности ограничения доступа к строке
------------------------------------------------------------
------------------------------------------------------------
-- 1. Проверка ограничения создания строки
------------------------------------------------------------
-- Подключаемся суперпользователем
\c - postgres
-- Выдаем права на чтение, изменение и удаление строк пользователю u2_0_00
GRANT SELECT, UPDATE, DELETE ON TABLE test1 TO u2_0_00;
-- Добавляем строку для проверки
INSERT INTO test1 VALUES (5,'create test', 'create test');
-- Подключаемся пользователем u2_0_00
\c - u2_0_00
-- Ожидается запрет создания строк
INSERT INTO test1 VALUES (6,'create test', 'create test');
-- Ожидается доступ на чтение ко всем строкам
SELECT * FROM test1;
-- Ожидается возможность модификации строк
UPDATE test1 SET for_all_value = 'test_value' WHERE for_all_key = 5;
-- Ожидается возможность удаления строк
DELETE FROM test1 WHERE for_all_key = 5;
------------------------------------------------------------
-- 2. Подготовка таблицы test1 к дальнейшим проверкам
------------------------------------------------------------
-- Подключаемся суперпользователем
\c - postgres
-- Добавляем колонку для проверки доступа
ALTER TABLE test1 ADD COLUMN allowed_user text;
-- Разрешаем всем доступ к новой колонке
GRANT SELECT (allowed_user) ON TABLE test1 TO public;
-- Включаем разграничение доступа на строках
ALTER TABLE test1 ENABLE ROW LEVEL SECURITY;
-- Создаем правило ограничения
CREATE POLICY heap_restrict ON test1 AS PERMISSIVE FOR ALL TO PUBLIC USING ( allowed_user = CURRENT_USER );
-- Разрешаем доступ пользователю u_0_00 к строкам с идентификаторами (for_all_key) 1, 2 и 3
UPDATE test1 SET allowed_user = 'u_0_00' WHERE for_all_key = ANY (ARRAY[1, 2, 3]);
-- Разрешаем доступ пользователю u1_0_00 к строкам с идентификатором (for_all_key) 4
UPDATE test1 SET allowed_user = 'u1_0_00' WHERE for_all_key = ANY (ARRAY[4]);
-- Проверяем содержимое таблицы. Для суперпользователя ожидается вывод всех строк.
SELECT * FROM test1;
------------------------------------------------------------
-- 3. Проверка ограничения доступа на чтение к строке
------------------------------------------------------------
-- Подключаемся пользователем u_0_00
\c - u_0_00
-- Проверяем доступ. Ожидается вывод трех строк с идентификаторами (for_all_key) 1, 2 и 3
SELECT * FROM test1;
-- Подключаемся пользователем u1_0_00
\c - u1_0_00
-- Проверяем доступ. Ожидается вывод одной строки с идентификатором (for_all_key) 4
SELECT for_all_key, for_all_value FROM test1;
------------------------------------------------------------
-- 4. Проверка ограничения модификации строки
------------------------------------------------------------
-- Подключаемся пользователем u1_0_00
\c - u1_0_00
-- Ожидается невозможность модификации строки с идентификаторм 1 (В выводе "UPDATE 0")
UPDATE test1 SET for_all_value = 'test_value' WHERE for_all_key = 1;
------------------------------------------------------------
-- 5. Проверка ограничения удаления строки
------------------------------------------------------------
-- Подключаемся пользователем u2_0_00
\c - u2_0_00
-- Ожидается невозможность удаления строки с идентификаторм 1 (В выводе "DELETE 0")
DELETE FROM test1 WHERE for_all_key = 1;
------------------------------------------------------------
-- III Проверка возможности ограничения доступа к строке
-- Выше была продемонстрирована возможность ограничения доступа к колонке и строке. Доступ к полю (ячейке) ограничивается посредством одновременного ограничения к колонке и строке.
------------------------------------------------------------
-- Очистка после теста
\c - postgres
DROP TABLE test1;
--
\i 'support/done.sql'
