-- RBT-TEST - МПРД при логической репликации
--
\set ECHO queries
-- Тестируется:
-- - Репликация таблицы с меткой на записях
-- - Репликация таблицы без метки на записях
-- - Репликация вставки записи с меткой
-- - Репликация вставки записи без метки
-- - Репликация удаления записи с меткой
-- - Репликация удаления записи без метки
-- - Репликация изменения записи с меткой
-- - Репликация изменения записи без метки
-- - Репликация изменения метки у записи
-- - Репликация данных между БД на одном кластере
-- - Репликация данных между БД двух кластеров
-- - Создание слота репликации отдельно от подписки и таблиц с меткой БД публикации
-- - Создание слота репликации в ответ на создание подписки с меткой БД публикации (libpq)
-- - Доступ к данным слота репликации пользователей с достаточным и недостаточным уровнем безопасности
-- - Соответствие меток созданных подписок меткам сессии
-- - Соответствие меток созданных публикаций меткам сессии
-- - Использование роли {2,3} для репликации
-- - Использование роли {0,0} для репликации
-- - Использование роли не postgres для создания подписки
-- - Изменение меток публикации, репликации и слота
-- - Ограничение максимального уровня метки слота
-- - Проверка совпадения метки таблицы публикации и метки таблицы подписки
--
-- TODO:
-- - Роль с недостаточными привилегиями для публикации данных безостановочно сообщает в лог ошибку
-- - Проверка метки {255, 0xFFFFFFFFFFFFFFFF}
-- ------------------------------------------------
--
-- Тестирование выполняется на одном кластере и на двух кластерах
--
-- База данных "mtest" основного кластера из support/init.sql публикует данные
-- База данных "replica" основного кластера и "replica" второго кластера принимают данные
--
-- Для публикации используются роли "u_2_11" {2,3} и u_0_00 {0,0}
--
-- Публикации "pbtt1" и "pbtt2" в "mtest" отслеживают все допустимые изменения в таблицах "Проверка" и "ПроверкаБезМеток" 
--
-- Подписка "suptt1" в "replica" настраивается на публикующий cлот "slot1". Полученные данные сохраняются в таблице "Проверка"
-- Подписка "suptt2" в "replica" настраивается на публикующий cлот "slot2". Полученные данные сохраняются в таблице "ПроверкаБезМеток"
-- Подписка "suptt1_II" второго кластера в "replica" настраивается на публикующий cлот "suptt1_II". Полученные данные сохраняются в таблице "Проверка"
-- Подписка "suptt2_II" второго кластера в "replica" настраивается на публикующий cлот "suptt2_II". Полученные данные сохраняются в таблице "ПроверкаБезМеток"
-- 
-- В postgresql.conf опция wal_level должна иметь значение logical
-- В версии Postgresql 11 в случае использования одного кластера требуется создавать подписку с условием create_slot = false
--
\set ECHO all
-- ------------------------------------------------
-- Настройка публикации
--
-- Создание БД, таблицы и роли, установка меток и признаков на внешние контейнеры
\set ECHO none
\i 'support/init.sql'
\i 'support/init-mac.sql'
\i 'support/create-test-table.sql'
\i 'support/support_functions.sql'
\set ECHO all
--
CREATE TABLE "ПроверкаБезМеток" (LIKE "Проверка");
ALTER TABLE "ПроверкаБезМеток" REPLICA IDENTITY FULL;
CREATE TABLE "ПроверкаРавенстваМетокТаб" (LIKE "Проверка");
ALTER TABLE "ПроверкаРавенстваМетокТаб" REPLICA IDENTITY FULL;
--
-- Роль соединения должна иметь возможность прочитать таблицы
ALTER TABLE "Проверка" OWNER TO u_2_11;
ALTER TABLE "ПроверкаРавенстваМетокТаб" OWNER TO u_2_11;
ALTER TABLE "ПроверкаБезМеток" OWNER TO u_0_00;
--
-- Проверяем тип репликации
show wal_level;
-- Выдаем роли соединения право на репликацию
ALTER ROLE u_2_11 REPLICATION;
ALTER ROLE u_0_00 REPLICATION;
ALTER ROLE u_1_01 REPLICATION;
ALTER ROLE u_1_01 WITH SUPERUSER;
--
-- Поднимаем метку таблицы до уровня роли
MAC LABEL ON TABLE "Проверка" is '{2,3}';
MAC LABEL ON TABLE "ПроверкаРавенстваМетокТаб" is '{2,3}';
--
-- Добавляем в таблицы данные для проверки первоначальной синхронизации
SET ac_session_maclabel='{0,0}'; INSERT INTO "ПроверкаБезМеток" VALUES (50, 60);
SET ac_session_maclabel='{1,3}'; INSERT INTO "Проверка" VALUES (101, 91);
SET ac_session_maclabel='{0,2}'; INSERT INTO "Проверка" VALUES (102, 92);
SET ac_session_maclabel='{0,2}'; INSERT INTO "ПроверкаРавенстваМетокТаб" VALUES (102, 92);
--
-- Создаем публикации
SET ac_session_maclabel='{0,2}'; CREATE PUBLICATION pbtt1 FOR TABLE "Проверка";
SET ac_session_maclabel='{0,2}'; CREATE PUBLICATION pbtt1error FOR TABLE "Проверка", "ПроверкаРавенстваМетокТаб";
SET ac_session_maclabel='{0,0}'; CREATE PUBLICATION pbtt2 FOR TABLE "ПроверкаБезМеток";
--
-- Создаем слот для подписки
-- Уровни меток равны уровню БД, но не сессии
SET ac_session_maclabel='{2,3}'; \copy (SELECT * FROM pg_create_logical_replication_slot('slot1', 'pgoutput')) to /dev/null
SET ac_session_maclabel='{0,0}'; \copy (SELECT * FROM pg_create_logical_replication_slot('slot2', 'pgoutput')) to /dev/null
--
-- Проверяем изменение метки слота до репликации
MAC LABEL ON slot slot1 IS '{2,2}';
SELECT slot_name, maclabel FROM pg_replication_slots order by slot_name;
-- ------------------------------------------------
-- Настройка подписки на основном кластере
--
-- Создаем базу и таблицы для реплики
CREATE DATABASE replica;
\connect replica postgres - 6000
\i 'support/init-replica.sql'
--
-- Создаем подписки
\connect replica u_1_01 - 6000
CREATE SUBSCRIPTION suptt1 CONNECTION 'host=localhost port=6000 dbname=mtest user=u_2_11' PUBLICATION pbtt1 WITH (create_slot = false, enabled = false);
\connect replica postgres - 6000
SET ac_session_maclabel='{0,0}'; CREATE SUBSCRIPTION suptt2 CONNECTION 'host=localhost port=6000 dbname=mtest user=postgres' PUBLICATION pbtt2 WITH (create_slot = false, enabled = false);
-- Попытка создать подписку на втором кластере для таблицы с иной меткой. Ожидается ошибка
\! psql -d replica -U postgres -p 6000 -c "CREATE SUBSCRIPTION suptt1error CONNECTION 'host=localhost port=6000 dbname=mtest user=u_2_11' PUBLICATION pbtt1error WITH (create_slot = false, enabled = false);"
--
-- Включаем подписки
ALTER SUBSCRIPTION suptt1 SET (slot_name='slot1');
ALTER SUBSCRIPTION suptt2 SET (slot_name='slot2');
ALTER SUBSCRIPTION suptt1 ENABLE;
ALTER SUBSCRIPTION suptt2 ENABLE;
--
-- Проверяем изменение метки подписки
MAC LABEL ON SUBSCRIPTION suptt1 IS '{2,0}';
SELECT subname, maclabel FROM PG_SUBSCRIPTION  order by subname;
--
\! sleep 2
-- ------------------------------------------------
-- Настройка подписки на втором кластере
--
\connect template1 postgres - 6001
-- Создаем базу для реплики на втором кластере
CREATE DATABASE replica;
\connect replica postgres - 6001
\i 'support/init-replica.sql'
--
-- Создаем подписки на втором кластере
SET client_min_messages TO warning;
SET ac_session_maclabel='{1,0}'; CREATE subscription suptt1_II connection 'host=localhost port=6000 dbname=mtest user=u_2_11' publication pbtt1;
CREATE subscription suptt2_II connection 'host=localhost port=6000 dbname=mtest user=u_0_00' publication pbtt2;
-- Попытка создать подписку на втором кластере для таблицы с иной меткой. Ожидается ошибка
\! psql -d replica -U postgres -p 6001 -c "CREATE subscription suptt1_IIerror connection 'host=localhost port=6000 dbname=mtest user=u_2_11' publication pbtt1error;"
SET client_min_messages TO notice;
--
-- Проверяем изменение метки подписки
MAC LABEL ON SUBSCRIPTION suptt1_II IS '{2,0}';
SELECT subname, maclabel FROM PG_SUBSCRIPTION order by subname;
-- ------------------------------------------------
-- Тестируем репликацию
\! sleep 2
--
-- Подключаемся к публикующей базе и проверяем метку слотов
-- Пользователь u_2_11 {2,3}
\connect mtest u_2_11 - 6000
--
SELECT * FROM advanced_waiting(query := 'SELECT COUNT(*) FROM pg_replication_slots', expected_result := 4, label := 'проверяем метку слотов - пользователь u_2_11 {2,3}');
--
SELECT slot_name, maclabel FROM pg_replication_slots order by slot_name;
-- Пользователь u_0_00 {0,0}
\connect mtest u_0_00 - 6000
--
SELECT * FROM advanced_waiting(query := 'SELECT COUNT(*) FROM pg_replication_slots', expected_result := 0, label := 'проверяем метку слотов - пользователь u_0_00 {0,0}');
--
SELECT slot_name, maclabel FROM pg_replication_slots order by slot_name;
--
-- Проверяем метки публикации
\connect mtest postgres - 6000
mac label on publication pbtt1 is '{2,3}';
mac label on publication pbtt2 is '{2,3}';
SELECT pubname, maclabel FROM pg_publication order by pubname;
-- Проверяем изменение метки слота после начала репликации
MAC LABEL ON slot slot1 IS '{1,1}';
MAC LABEL ON slot slot2 IS '{1,1}';
-- Проверяем ограничение изменения метки слота. Ожидается ошибка
\! psql -d mtest -U postgres -p 6000 -c "MAC LABEL ON slot slot2 IS '{10,10}';"
--
SELECT * FROM advanced_waiting(query := 'SELECT COUNT(*) FROM pg_replication_slots', expected_result := 4, label := 'Проверяем ограничение изменения метки слота');
--
SELECT slot_name, maclabel FROM pg_replication_slots order by slot_name;
-- Добавляем строки
SET ac_session_maclabel='{1,0}'; INSERT INTO "Проверка" VALUES (103, 93);
SET ac_session_maclabel='{2,1}'; INSERT INTO "Проверка" VALUES (104, 94);
SET ac_session_maclabel='{2,3}'; INSERT INTO "Проверка" VALUES (105, 95);
SET ac_session_maclabel='{2,3}'; INSERT INTO "Проверка" VALUES (106, 'удаление');
SET ac_session_maclabel='{2,3}'; INSERT INTO "Проверка" VALUES (107, 'обновление');
SET ac_session_maclabel='{2,3}'; INSERT INTO "Проверка" VALUES (108, 'метка {2,2}');
SET ac_session_maclabel='{0,0}'; INSERT INTO "ПроверкаБезМеток" VALUES (20, 80);
SET ac_session_maclabel='{0,0}'; INSERT INTO "ПроверкаБезМеток" VALUES (30, 'удаление');
SET ac_session_maclabel='{0,0}'; INSERT INTO "ПроверкаБезМеток" VALUES (40, 'обновление');
--
\! sleep 1
--
\connect mtest postgres - 6000
--
-- Удаление записи
DELETE FROM "Проверка" WHERE "идентификатор"=106;
DELETE FROM "ПроверкаБезМеток" WHERE "идентификатор"=30;
--
-- Обновление записи
UPDATE "Проверка" SET "данные"='обновлено' WHERE "идентификатор"=107;
UPDATE "ПроверкаБезМеток" SET "данные"='обновлено' WHERE "идентификатор"=40;
--
-- Изменение метки записи
CHMAC "Проверка" SET maclabel='{2,2}' WHERE "идентификатор"=108;
--
-- Проверяем таблицы в публикующей базе
SELECT "идентификатор", "данные", maclabel FROM "Проверка" order by "идентификатор";
--
SELECT * FROM advanced_waiting(query := 'SELECT COUNT(*) FROM "ПроверкаБезМеток"', expected_result := 3, label := 'Проверяем таблицы в публикующей базе');
--
SELECT "идентификатор", "данные" FROM "ПроверкаБезМеток" order by "идентификатор";
--
-- Подключаемся к реплике
\connect replica postgres - 6000
-- Проверяем метки подписок
SELECT * FROM advanced_waiting(query := 'SELECT COUNT(*) FROM pg_subscription', expected_result := 2, label := 'Проверяем метки подписок');
--
SELECT subname, maclabel FROM pg_subscription order by subname;
--
-- Проверяем таблицы в реплике
SELECT * FROM advanced_waiting(query := 'SELECT COUNT(*) FROM "Проверка"', expected_result := 7, label := 'Проверяем все таблицы в реплике');
--
SELECT "идентификатор", "данные", maclabel FROM "Проверка" order by "идентификатор";
--
SELECT * FROM advanced_waiting(query := 'SELECT COUNT(*) FROM "ПроверкаБезМеток"', expected_result := 3, label := 'Проверяем таблицы без меток в реплике');
--
SELECT "идентификатор", "данные" FROM "ПроверкаБезМеток" order by "идентификатор";
--
-- Подключаемся к реплике на втором кластере
\connect replica postgres - 6001
-- Проверяем таблицы реплики второго кластера
SELECT * FROM advanced_waiting(query := 'SELECT COUNT(*) FROM "Проверка"', expected_result := 7, label := 'Проверяем все таблицы реплики второго кластера');
--
SELECT "идентификатор", "данные", maclabel FROM "Проверка" order by "идентификатор";
--
SELECT * FROM advanced_waiting(query := 'SELECT COUNT(*) FROM "ПроверкаБезМеток"', expected_result := 3, label := 'Проверяем таблицы без меток реплики второго кластера');
--
SELECT "идентификатор", "данные" FROM "ПроверкаБезМеток" order by "идентификатор";
-- ------------------------------------------------
-- Завершаем тест
-- Выключаем всё на реплике второго кластера и сбрасываем мандатные атрибуты кластера
\connect replica postgres - 6001
ALTER SUBSCRIPTION suptt1_II DISABLE;
ALTER SUBSCRIPTION suptt2_II DISABLE;
SET client_min_messages TO WARNING;
DROP SUBSCRIPTION IF EXISTS suptt1_II;
DROP SUBSCRIPTION IF EXISTS suptt2_II;
SET client_min_messages TO NOTICE;
\connect template1 postgres - 6001
DROP DATABASE replica;
--
\set ECHO errors
\i 'support/reset-cluster-mac-attributes.sql'
\set ECHO all
--
-- Выключаем всё на реплике основного кластера
\connect replica postgres - 6000
ALTER SUBSCRIPTION suptt1 DISABLE;
ALTER SUBSCRIPTION suptt2 DISABLE;
ALTER SUBSCRIPTION suptt1 SET (slot_name=NONE);
ALTER SUBSCRIPTION suptt2 SET (slot_name=NONE);
DROP SUBSCRIPTION IF EXISTS suptt1;
DROP SUBSCRIPTION IF EXISTS suptt2;
--
-- Выключаем всё на публикаторе основного кластера
\connect mtest postgres
SELECT pg_drop_replication_slot(slot_name) FROM pg_replication_slots WHERE slot_name = 'slot1';
SELECT pg_drop_replication_slot(slot_name) FROM pg_replication_slots WHERE slot_name = 'slot2';
--
-- Удаляем БД реплики и завершаем работу с кластером
DROP DATABASE replica;
--
\set ECHO errors
\i 'support/done.sql'
\i 'support/reset-cluster-mac-attributes.sql'
\set ECHO all
