import argparse as ap
from os import environ

DEFAULT = '. (По умолчанию "%(default)s")'
PG_VERSION = environ.get('PG_VERSION', 15)
SCRIPT_VERSION = environ.get('SCRIPT_VERSION', '$SCRIPT_VERSION')

environ["COLUMNS"] = "150"  # сделаем сообщения помощи от argparse более красивыми


class ArgumentConstsHelpFormatter(ap.ArgumentDefaultsHelpFormatter):
    def _get_help_string(self, action):
        _help = action.help
        if _help is None:
            _help = ''

        if (action.const is not None and type(action.const) is not bool) and \
           ('%(const)' not in _help) and ('По умолчанию' not in _help):
            defaulting_nargs = [ap.OPTIONAL, ap.ZERO_OR_MORE]

            if (action.option_strings and action.option_strings[0] not in ('-h', '--version')) \
               or action.nargs in defaulting_nargs:
                _help += f'. (По умолчанию "%(const)s").'

        return _help


class SplitStoreAction(ap._StoreAction):
    split_by = ','

    @classmethod
    def set_split_by(cls, split_by):
        cls.split_by = split_by

    def __call__(self, parser, namespace, values, option_string=None):
        if type(values) is str:
            values = values.split(SplitStoreAction.split_by)
        setattr(namespace, self.dest, values)


def __additionally_parse_arguments(args: ap.Namespace):
    if getattr(args, 'type', None) is not None:
        args.type = ' '.join(args.type)
    else:
        args.type = 'all'  # если не был задан флаг для выбора тестов - по умолчанию запускаем все тесты

    # Обработка "исключений"
    if getattr(args, 'working_dir', None) is not None:
        args.working_dir = f"$(realpath -eq '{args.working_dir}' || echo '{args.working_dir}')"

    # Цикл по атрибутам объекта args
    for attr, value in ((a, getattr(args, a)) for a in dir(args) if not a.startswith('_')):
        if type(value) == list:
            setattr(args, attr, ' '.join(value))


def print_bash_variables(args: ap.Namespace, cmd: str):
    __additionally_parse_arguments(args)

    print(f'ARG_CMD={cmd}')
    for arg in vars(args):
        value = getattr(args, arg)
        if type(value) is bool:
            print(f'ARG_{arg.upper()}={int(value)}')
        else:
            print(f'ARG_{arg.upper()}="{value}"')


def main():
    parser = ap.ArgumentParser(
        prog='runtests',
        description="Скрипт для запуска se-тестов (требуются root права).",
        formatter_class=ArgumentConstsHelpFormatter,
        argument_default=ap.SUPPRESS,
        allow_abbrev=False,
        add_help=False  # добавим самостоятельно ниже
    )
    parser.add_argument('-h', '--help', action='help', help='Печать данной справки и выход')
    parser.add_argument('--version', action='version', version=f'Версия скрипта: {SCRIPT_VERSION}',
                        help='Печать версии скрипта и выход')
    subparsers = parser.add_subparsers(title='Подкоманды', description='(в основном для разработчиков)',
                                       help='Доступные подкоманды (укажите подкоманду и флаг -h/--help для справки)',
                                       dest='subparser_name')

    ### Команда "main" (т.е. неуказание какой-либо команды, только параметры)
    g = parser.add_argument_group('Опции - тесты', 'Указание тестов для выполнения')
    tests = g.add_mutually_exclusive_group(required=False)
    other = parser.add_argument_group('Опции - разное', 'Различные другие флаги')
    dev = parser.add_argument_group('Опции - DEV', 'Параметры для разработчиков')

    tests.add_argument('-f', '--file', help='Указанный файл теста')
    tests.add_argument('-l', '--list', nargs='?', const='./tests.lst',
                         help=f'Файл со списком тестов')
    tests.add_argument('-t', '--type', nargs='?', const=['all'], action=SplitStoreAction,
                         help=f'Указанные (через запятую) типы или все тесты [all,acl,mac,misc]')

    other.add_argument('-v', '--pg-version', help=f'Указать версию PostgreSQL')

    dev.add_argument('-d', '--working-dir',
                     default=f'/usr/share/postgresql/{PG_VERSION}/test/pgacext',
                     help=f'Указать директорию для запуска тестов (принимается и относительный путь){DEFAULT}')
    dev.add_argument('-n', '--no-rm', action='store_true', help='Не удалять кластеры после тестов')
    dev.add_argument('-N', '--no-stop', action='store_true', help='Не останавливать кластеры после тестов')
    dev.add_argument('-u', '--no-rm-users', action='store_true', help='Не удалять пользователей после тестов')
    dev.add_argument('--rr', nargs='?', const=['setest'], action=SplitStoreAction,
                     help='Запуск указанных (через запятую) кластеров под реверсивным отладчиком [rr]')
    dev.add_argument('--clusters', action=SplitStoreAction,
                     help='Запуск только указанных (через запятую) кластеров. По умолчанию запускаются все кластеры')
    ###

    ### Подкоманды
    subcommands = (
        # подкоманда, её синонимы, описание
        ('prepare_environment', ['pe'], 'Создание окружения для тестов (пользователи, кластеры и др.). '
                                        'Можно задавать некоторые параметры (перед командой) как при запуске тестов'),
        ('clean_environment', ['ce'], 'Очистка окружения. Можно задавать некоторые параметры (перед командой) как при запуске тестов'),
        ('create_users', ['cu'], 'Создание тестовых пользователей'),
        ('remove_users', ['ru'], 'Удаление тестовых пользователей'),
        ('variables', ['vars'],
         '''Выставление shell переменных (пример использования - 'eval "$(./runtests vars)"; echo $PG_WORKDIR')'''),
    )
    for cmd, aliases, desc in subcommands:
        sub_parser = subparsers.add_parser(cmd, description=f'{desc}.', aliases=aliases, add_help=False)
        sub_parser.add_argument('-h', '--help', action='help', help='Печать данной справки и выход')
    ###

    # Непосредственный парсинг аргументов
    args = parser.parse_args()
    if args.subparser_name is None:
        args.subparser_name = 'main'

    print_bash_variables(args, args.subparser_name)


if __name__ == '__main__':
    main()
