CREATE EXTENSION IF NOT EXISTS plpython3u;
CREATE EXTENSION IF NOT EXISTS pageinspect;
--
-- From https://wiki.postgresql.org/wiki/Index_Maintenance:
-- "Note: The byteswap is only necessary on little-endian (Intel) CPUs."
--
CREATE OR REPLACE FUNCTION reverse_bytea(input_array bytea)
    RETURNS bytea stable
    LANGUAGE plpython3u AS
$$
    b = bytearray()
    b.extend(input_array)
    b.reverse()
    return b
$$;
--
CREATE OR REPLACE FUNCTION get_columns_storage_type(_table text)
    RETURNS TABLE (n int2, column_name name, storage_type char)
    LANGUAGE plpgsql AS
$$
BEGIN
    RETURN QUERY
    EXECUTE format('
        SELECT a.attnum, a.attname, t.typstorage::char
        FROM
            pg_class c
            JOIN pg_attribute a ON a.attrelid = c.oid
            JOIN pg_type t ON a.atttypid = t.oid
        WHERE c.relname = ''%s'' AND not a.attisdropped;
    ',
    _table);
END;
$$;
--
--
--
CREATE OR REPLACE FUNCTION raw_table_data(table_name TEXT)
	RETURNS TABLE (
		lp smallint, lp_off smallint, lp_flags smallint, lp_len smallint, t_xmin xid, t_xmax xid,
		t_field3 integer, t_ctid tid, t_infomask2 integer, t_infomask integer, t_hoff smallint,
		t_bits text, t_oid oid, t_data bytea, t_infomaskpgac integer, t_hasmac boolean, t_maclabel text,
		--
		ctid tid, blkno bigint, column_name name, column_num int2, storage_type char, chunk_id int,
        --
        xmax_is_lock boolean, xmax_committed boolean, xmax_rolled_back boolean, xmax_multixact boolean,
        --
        t_ccr_is_on bool
	)
    LANGUAGE plpgsql AS
$$
DECLARE
    page_size INT;
BEGIN
    SET client_min_messages TO warning;
    page_size = (SELECT current_setting('block_size'));
--
	RETURN QUERY
	SELECT
		hpi.*, (_blkno,hpi.lp)::text::tid AS ctid, _blkno, col.column_name, col.n, col.storage_type,
        substr(reverse_bytea(substr(hpia.t_attrs[col.n], length(hpia.t_attrs[col.n])-7, 4))::text, 2)::bit(32)::int AS chunk_id,
        (hpi.t_infomask & 128)::boolean AS xmax_is_lock, (hpi.t_infomask & 1024)::boolean AS xmax_committed,
        (hpi.t_infomask & 2048)::boolean AS xmax_rolled_back, (hpi.t_infomask & 4096)::boolean AS xmax_multixact,
        ((hpi.t_infomaskpgac & 4) = 0) AS "CCR is ON"
    FROM
        (SELECT * FROM get_columns_storage_type(table_name) a WHERE a.storage_type IN ('x', 'e')) AS col,
        generate_series(0, pg_relation_size(table_name::regclass::text) / page_size - 1) AS _blkno,
        heap_page_item_attrs(get_raw_page(table_name, _blkno::int), table_name::regclass) AS hpia
		JOIN heap_page_items(get_raw_page(table_name, _blkno::int)) AS hpi ON hpia.lp_off = hpi.lp_off
    WHERE
        substr(hpia.t_attrs[col.n]::text, 3, 2) = '01';
--
    RESET client_min_messages;
END
$$;
-- SELECT * FROM raw_table_data('docs');
--
--
--
CREATE OR REPLACE FUNCTION raw_toast_data(toast_table_name TEXT)
	RETURNS TABLE (
		lp smallint, lp_off smallint, lp_flags smallint, lp_len smallint, t_xmin xid, t_xmax xid,
		t_field3 integer, t_ctid tid, t_infomask2 integer, t_infomask integer, t_hoff smallint,
		t_bits text, t_oid oid, t_data bytea, t_infomaskpgac integer, t_hasmac boolean, t_maclabel text,
		--
		ctid tid, chunk_id oid, chunk_seq int, chunk_data bytea,
		t_ccr_is_on bool, blkno bigint
	)
    LANGUAGE plpgsql AS
$$
DECLARE
    page_size INT;
BEGIN
    SET client_min_messages TO warning;
    page_size = (SELECT current_setting('block_size'));
--
	RETURN QUERY
	EXECUTE format(
		'SELECT hpi.*, (_blkno,hpi.lp)::text::tid AS ctid, toast.*, ((t_infomaskpgac & 4) = 0) AS "CCR is ON", _blkno
		 FROM
		 	generate_series(0, pg_relation_size(''pg_toast.%I''::regclass::text) / $1 - 1) AS _blkno,
		 	pg_toast.%I AS toast
		 		JOIN heap_page_items(get_raw_page(''pg_toast.%I'', _blkno::int)) AS hpi ON ctid = t_ctid',
		toast_table_name, toast_table_name, toast_table_name
	)
    USING page_size;
--
    RESET client_min_messages;
END
$$;
-- SELECT * FROM raw_toast_data('pg_toast_2618');
--
--
--
CREATE OR REPLACE FUNCTION raw_toast_index_data(index_table_name TEXT)
	RETURNS TABLE (
        itemoffset smallint, ctid tid, itemlen smallint, nulls boolean, vars boolean,
        data text, dead boolean, htid tid, tids tid[], t_hasmac boolean, t_maclabel text,
        --
		blkno bigint, chunk_id int, chunk_seq int
	)
	LANGUAGE plpgsql AS
$$
DECLARE
    page_size INT;
BEGIN
    SET client_min_messages TO warning;
    page_size = (SELECT current_setting('block_size'));
--
	RETURN QUERY
	SELECT
		pi.*, _blkno,
		(('x' || regexp_replace(substr(pi.data,1,11), '([^ ]*) ([^ ]*) ([^ ]*) ([^ ]*)', '\4\3\2\1'))::bit(32)::int) AS chunk_id,
		(('x' || regexp_replace(substr(pi.data,13,23), '([^ ]*) ([^ ]*) ([^ ]*) ([^ ]*)', '\4\3\2\1'))::bit(32)::int) AS chunk_seq
	FROM
		generate_series(1, pg_relation_size(('pg_toast.' || index_table_name)::regclass::text) / page_size - 1) AS _blkno,
		bt_page_items(('pg_toast.' || index_table_name)::regclass::text, _blkno::int) AS pi;
-- 	WHERE pi.data not like ''
--
    RESET client_min_messages;
END
$$;
-- SELECT * FROM raw_toast_index_data('pg_toast_2618_index');
--
--
--
CREATE OR REPLACE FUNCTION live_toast_index_pages(index_table_name TEXT)
	RETURNS TABLE (
	    blkno bigint, type "char", live_items integer, dead_items integer, avg_item_size integer, page_size integer,
        free_size integer, btpo_prev bigint, btpo_next bigint, btpo_level bigint, btpo_flags integer
    )
	LANGUAGE plpgsql AS
$$
DECLARE
    page_size INT;
BEGIN
    SET client_min_messages TO warning;
    page_size = (SELECT current_setting('block_size'));
--
	RETURN QUERY
	SELECT ps.*
	FROM
		generate_series(1, pg_relation_size(('pg_toast.' || index_table_name)::regclass::text) / page_size - 1) AS _blkno,
		bt_page_stats(('pg_toast.' || index_table_name)::regclass::text, _blkno::int) AS ps
	WHERE
		ps.live_items != 0 AND ps.btpo_flags != 0 AND ps.type NOT IN ('d', 'i', 'e');
--
    RESET client_min_messages;
END
$$;
-- SELECT * FROM live_toast_index_pages('pg_toast_2618_index');
--
--
--
CREATE OR REPLACE FUNCTION ctid_from_var_table(schema text, table_name text)
    RETURNS TABLE(ctid tid)
    LANGUAGE plpgsql AS
$$
BEGIN
    IF schema != '' THEN
        RETURN QUERY EXECUTE format('SELECT ctid FROM %I.%I', schema, table_name);
    ELSE
        RETURN QUERY EXECUTE format('SELECT ctid FROM %I', table_name);
    END IF;
END
$$;
--
--
--
-- Выдержка из 'src/include/access/nbtree.h':
-- #define P_RIGHTMOST(opaque)		((opaque)->btpo_next == P_NONE)
-- *	On a non-rightmost page, the high key lives in item 1 and data items
-- *	start in item 2.  Rightmost pages have no high key, so we store data
-- *	items beginning in item 1.
-- #define P_HIKEY				    ((OffsetNumber) 1)
-- #define P_FIRSTKEY			    ((OffsetNumber) 2)
-- #define P_FIRSTDATAKEY(opaque)	(P_RIGHTMOST(opaque) ? P_HIKEY : P_FIRSTKEY)
--
CREATE OR REPLACE FUNCTION get_toast_and_its_index_mac_attrs(toasted_table_name TEXT)
    RETURNS TABLE (
		"table" text, ctid tid, chunk_id int, chunk_seq int,
		t_hasmac bool, t_maclabel text, t_ccr_is_on bool, count_unique bigint
	)
    LANGUAGE plpgsql AS
$$
DECLARE
    toast_name pg_class.relname%TYPE;
    index_toast_name pg_class.relname%TYPE;
BEGIN
    toast_name = (SELECT relname FROM pg_class WHERE oid =
                  (SELECT reltoastrelid FROM pg_class WHERE relname=toasted_table_name));
    index_toast_name = (SELECT format('%s_index', toast_name));
--
	RETURN QUERY
--
	WITH
    chunk_ids AS (
		SELECT c.chunk_id
		FROM raw_table_data(toasted_table_name) c
        JOIN ctid_from_var_table('', toasted_table_name) alive ON alive.ctid = c.ctid  -- живые записи
	)
--
	-- МРД атрибуты записи из toast таблицы
	SELECT 'toast', a.t_ctid, a.chunk_id::int, a.chunk_seq, a.t_hasmac, a.t_maclabel, a.t_ccr_is_on, COUNT(*)
	FROM
	    chunk_ids ci
	    JOIN raw_toast_data(toast_name) a ON ci.chunk_id = a.chunk_id
        JOIN ctid_from_var_table('pg_toast', toast_name) alive ON alive.ctid = a.ctid  -- живые записи
    GROUP BY (a.t_ctid, a.chunk_id, a.chunk_seq, a.t_hasmac, a.t_maclabel, a.t_ccr_is_on)
--
	UNION ALL
--
	-- МРД атрибуты записи из toast index таблицы (все страницы, кроме "крайней правой")
	SELECT 'toast_index', b.ctid, b.chunk_id, b.chunk_seq, b.t_hasmac, b.t_maclabel, null::boolean, COUNT(*)
	FROM
		live_toast_index_pages(index_toast_name) c
		JOIN raw_toast_index_data(index_toast_name) b ON c.blkno = b.blkno
		JOIN chunk_ids ci ON ci.chunk_id = b.chunk_id
	WHERE ((c.btpo_next != 0 AND b.itemoffset >= 2) OR c.btpo_next = 0) AND not b.dead
	GROUP BY (b.ctid, b.chunk_id, b.chunk_seq, b.t_hasmac, b.t_maclabel);
END
$$;
-- SELECT *
-- FROM get_toast_and_its_index_mac_attrs('pg_rewrite')
-- ORDER BY "table", t_maclabel, count_unique;
--
--
--
CREATE OR REPLACE FUNCTION maclabel_from_var_table(schema text, table_name text)
    RETURNS TABLE(maclabel maclabel)
    LANGUAGE plpgsql AS
$$
BEGIN
    IF schema != '' THEN
        RETURN QUERY EXECUTE format('SELECT maclabel FROM %I.%I', schema, table_name);
    ELSE
        RETURN QUERY EXECUTE format('SELECT maclabel FROM %I', table_name);
    END IF;
END
$$;
--
--
--
CREATE OR REPLACE VIEW system_tables
AS
SELECT c.*
FROM pg_class c
    JOIN pg_namespace n ON c.relnamespace = n.oid
    JOIN pg_attribute a ON a.attrelid = c.oid
WHERE c.relkind = 'r'
    AND a.attnum > 0
    AND NOT a.attisdropped
    AND n.nspname = 'pg_catalog'
    AND n.nspname != 'information_schema'
ORDER BY c.oid;
--
--
--
--
CREATE OR REPLACE VIEW system_tables_with_toasts
AS
WITH toasted_system_tables AS (
    SELECT DISTINCT s.oid, s.relname
    FROM system_tables s
        JOIN pg_attribute a ON a.attrelid = s.oid
        JOIN pg_type t ON a.atttypid = t.oid
    WHERE t.typstorage IN ('x', 'e')
    ORDER BY s.oid
)
SELECT tst.relname as toasted_table, c.relname as toast
FROM toasted_system_tables tst
JOIN pg_class c ON c.oid != tst.oid
WHERE c.relname like 'pg_toast_' || tst.oid || '%';
--
--
--
CREATE OR REPLACE VIEW get_system_tables_toast_mac_attrs
AS
WITH toasted_tables AS (
    SELECT DISTINCT v.toasted_table FROM system_tables_with_toasts v
)
SELECT DISTINCT tt.toasted_table, a."table",
    a.t_hasmac, a.t_maclabel, a.t_ccr_is_on, a.count_unique, COUNT(*) AS count_all
FROM toasted_tables tt
CROSS JOIN LATERAL get_toast_and_its_index_mac_attrs(tt.toasted_table) a
GROUP BY tt.toasted_table, a."table", a.t_hasmac, a.t_maclabel, a.t_ccr_is_on, a.count_unique
ORDER BY tt.toasted_table, a."table", a.t_maclabel, a.count_unique;
