import argparse
import jinja2
import os
import re
from glob import glob
from pathlib import Path


def read_file_without_empty_lines(filename):
    with open(filename) as file:
        return file.read().replace('\n\n', '\n').rstrip('\n')


def regex_replace(s, find, replace):
    return re.sub(find, replace, s)


def regex(s, find):
    res = re.search(find, s)
    if res is not None:
        return res[0]
    return ''


def mac_values_formatting(default, format_args, index, special=False):
    """
    format_args = [
        ('level,category', index), ...
    ]
    """
    i = 0
    if special:
        i = 1
    format_value = format_args[index][i] if len(format_args) else default

    value = '{'+format_value+'}'
    value = '{'+default+'}' if value == '{}' else value
    length = len(value)

    header_left_len = (length - 3) // 2
    header_right_len = header_left_len if length % 2 != 0 else header_left_len + 1
    header = ' '*header_left_len + 'MAC' + ' '*header_right_len

    separator = '-' * length

    return value, header, separator


class Template:
    def __init__(self, default_result_dir='expected'):
        env_folders = (
            'expected', 'templates', 'templates/expected', 'templates/sql',
            'templates/special_support', 'templates/support'
        )
        loader = jinja2.FileSystemLoader(env_folders)
        self.env = jinja2.Environment(loader=loader, lstrip_blocks=True, trim_blocks=True)

        self.env.globals['read_file_without_empty_lines'] = read_file_without_empty_lines
        self.env.globals['mac_values_formatting'] = mac_values_formatting
        self.env.filters['regex_replace'] = regex_replace
        self.env.filters['regex'] = regex

        self.env.globals['PG_VERSION'] = os.environ['PG_VERSION']

        self.env.globals['SPECIAL_SUPPORT_FILES'] = set()
        self.prepare_expected_templates_generation()

        self.default_result_dir = default_result_dir

    def prepare_expected_templates_generation(self):
        # 'Особенные' файлы support - это те, в которых есть 'SELECT' запросы или '\i' команды
        generated_dir = 'expected/generated_support'
        Path(generated_dir).mkdir(parents=True, exist_ok=True)

        for file in sorted(glob('templates/special_support/*.j2')):
            filename = os.path.basename(file.replace('.j2', ''))
            self.env.globals['SPECIAL_SUPPORT_FILES'].add(filename)
            self.create_file_from_template(file, result_dir=generated_dir)

    def create_file_from_template(self, template_path, result_dir=None):
        if result_dir is None:
            result_dir = self.default_result_dir
        filename = os.path.basename(template_path)
        template = self.env.get_template(filename)

        with open(f'{result_dir}/{filename.replace(".j2", "")}', 'w') as file:
            file.write(template.render())


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='command', help='commands')

    subparsers.add_parser('all')

    sub2 = subparsers.add_parser('one')
    sub2.add_argument('expected_template_path')

    args = parser.parse_args()

    template = Template()

    # Support файлы и SQL тесты
    for folder in ('support', 'sql'):
        for file in sorted(glob(f'templates/{folder}/*.j2')):
            template.create_file_from_template(file, result_dir=folder)

    # Файлы ожиданий
    if args.command == 'all':
        for file in sorted(glob('templates/expected/*.j2')):
            template.create_file_from_template(file)
    elif args.command == 'one':
        template.create_file_from_template(args.expected_template_path)


if __name__ == '__main__':
    main()
